#!/bin/python3

import argparse
import os
import json
import ROOT
import random
import numpy as np
from ROOT import gROOT, gInterpreter
from plot_utils import prepare_canvas, my_style, nthreads


if __name__ == "__main__":
    ROOT.ROOT.EnableImplicitMT(nthreads)
    # make sure the relative path for root loading is correct
    owd = os.getcwd()
    script_dir = os.path.dirname(os.path.realpath(__file__))
    os.chdir(os.path.join(script_dir, '..'))
    print('working directory is ' + os.getcwd())

    # get layouts
    with open(os.path.join(script_dir, '..', 'database', 'channels_layout.json'), 'r') as f:
        channels_layout = json.load(f)
    # format keys
    layout = {key.replace(' ', '_').upper(): channels for key, channels in channels_layout.items()}

    # argument parser
    parser = argparse.ArgumentParser('draw events')
    parser.add_argument('--min', dest='min', default=1, type=int, help='minimum number of hits')
    parser.add_argument('--max', dest='max', default=10, type=int, help='maximum number of hits')
    parser.add_argument('-o', '--plot-dir', dest='out_dir', default="./plots", help='output directory')
    parser.add_argument('-t', type=float, default=0.0, dest='thres', help='peak height threshold to sum')
    args = parser.parse_args()

    # recover paths
    for attr in ['out_dir']:
        setattr(args, attr, os.path.join(owd, getattr(args, attr)))

    os.makedirs(args.out_dir, exist_ok=True)

    # batch mode
    gROOT.SetBatch(True)
    # root macros
    gROOT.Macro('rootlogon.C')
    # declare root graph code
    func_code = open(os.path.join(script_dir, 'root', 'dataframe_analysis.C')).read()
    gInterpreter.Declare(func_code)

    # style
    my_style.cd()

    # root dataframes
    runs = [
        (
            'Low Rate Beam Test',
            {'SetLineColor': [30], 'SetFillColorAlpha': [30, 0.8], 'SetFillStyle': [3006]},
            ['processed_data/run420_0.root',],
            'unpack_integral(C2.peaks, true, 0, 1.0)',
            'true',
        ),
        (
            'High Rate Beam Test',
            {'SetLineColor': [38], 'SetFillColorAlpha': [38, 0.8], 'SetFillStyle': [3004]},
            ['processed_data/run308_0.root',],
            'unpack_integral(C2.peaks, true, 0, 1.0)',
            'true',
        ),
        (
            'Cosmic MIP Signal',
            {'SetLineColor': [46], 'SetFillColorAlpha': [46, 0.8], 'SetFillStyle': [3005]},
            ['processed_data/run233_0.root',],
            'unpack_integral(C2.peaks, true, 0, 1.0)',
            'true',
        ),
    ]
    '''
    runs = [
        (
            'Full Data',
            {'SetLineColor': [30], 'SetFillColorAlpha': [30, 0.8], 'SetFillStyle': [3006]},
            ['processed_data/run302_0.root',],
            'get_maximum_peak(C2.peaks).integral',
            'true'
        ),
        (
            'Cher. Cut Data',
            {'SetLineColor': [38], 'SetFillColorAlpha': [38, 0.8], 'SetFillStyle': [3004]},
            ['processed_data/run302_0.root',],
            'get_maximum_peak(C2.peaks).integral',
            'cher_cut_rough({'
                'Cer11_1,Cer11_2,Cer11_3,Cer11_4,'
                'Cer12_1,Cer12_2,Cer12_3,Cer12_4,'
                'Cer13_1,Cer13_2,Cer13_3,Cer13_4,'
                'Cer14_1,Cer14_2,Cer14_3,Cer14_4,'
                'Cer21_1,Cer21_2,Cer21_3,Cer21_4,'
                'Cer22_1,Cer22_2,Cer22_3,Cer22_4,'
                'Cer23_1,Cer23_2,Cer23_3,Cer23_4,'
                'Cer24_1,Cer24_2,Cer24_3,Cer24_4,'
                'Cer31_1,Cer31_2,Cer31_3,Cer31_4,'
                'Cer32_1,Cer32_2,Cer32_3,Cer32_4,'
                'Cer33_1,Cer33_2,Cer33_3,Cer33_4,'
                'Cer34_1,Cer34_2,Cer34_3,Cer34_4,'
                'Cer41_1,Cer41_2,Cer41_3,Cer41_4,'
                'Cer42_1,Cer42_2,Cer42_3,Cer42_4,'
                'Cer43_1,Cer43_2,Cer43_3,Cer43_4,'
                'Cer44_1,Cer44_2,Cer44_3,Cer44_4'
            '})'
        ),
    ]
    '''
    '''
    runs = [
        (
            'Low Rate Beam Test',
            {'SetLineColor': [30], 'SetFillColorAlpha': [30, 0.8], 'SetFillStyle': [3006]},
            ['processed_data/run420_0.root',],
            'get_maximum_peak(C2.peaks).height',
            'get_maximum_peak(C2.peaks).height > 0',
        ),
        (
            'High Rate Beam Test',
            {'SetLineColor': [46], 'SetFillColorAlpha': [46, 0.8], 'SetFillStyle': [3004]},
            ['processed_data/run308_0.root',],
            'get_maximum_peak(C2.peaks).height',
            'get_maximum_peak(C2.peaks).height > 0',
        ),
    ]
    '''
    xlbl, hrange = 'ADC Value', (500, 0, 20000)

    c = ROOT.TCanvas('Total Sum', '', 1920, 1280)
    pad = prepare_canvas(c, [[0]], 'Signal Sums Total', xlbl, 'Normalized Counts')[0]
    channels = ['{}_peak'.format(c) for chs in layout['LAPPD'] for c in chs]

    pad.cd()
    pad.SetLogy()
    hists = []
    hs = ROOT.THStack
    for i, (name, settings, root_files, func, fc) in enumerate(runs):
        chain = ROOT.TChain('EvTree')
        for rfile in root_files:
            chain.Add(rfile)
        rdf = ROOT.RDataFrame(chain)
        hist = rdf.Filter(fc).Define('temp', func).Histo1D(ROOT.RDF.TH1DModel(name, '', *hrange), 'temp')
        for method, inputs in settings.items():
            method_to_call = getattr(hist, method)
            method_to_call(*inputs)
        chist = hist.DrawNormalized('same')
        chist.SetMaximum(1.5)
    pad.BuildLegend(0.5, 0.8, 0.9, 0.9)
    c.SaveAs(os.path.join(args.out_dir, 'sum_comp.png'))

