import os
import argparse
import pandas as pd
import numpy as np
import ROOT


def load_root_macros(arg_macros):
    for path in arg_macros.split(','):
        path = path.strip()
        if os.path.exists(path):
            ROOT.gROOT.Macro(path)
        else:
            print('\"{}\" does not exist, skip loading it.'.format(path))


def max_peak_height(ch):
    if not len(ch.peaks):
        return 0
    return np.max([p.height for p in ch.peaks])



if __name__ == '__main__':
    parser = argparse.ArgumentParser()

    parser.add_argument(
            'file', type=str,
            help='path to root file.'
            )
    parser.add_argument(
            '-o', type=str,
            default='plots',
            dest='outdir',
            help='output directory.'
            )
    parser.add_argument(
            '-n', '--num-events', type=int,
            default=-1,
            help='number of events to process.'
            )
    parser.add_argument(
            '--start-event', type=int,
            default=0,
            help='start event.'
            )
    parser.add_argument(
            '--signal-cut', type=float,
            default=20,
            help='cut signal in height.'
            )

    args = parser.parse_args()

    load_root_macros('rootlogon.C')
    os.makedirs(args.outdir, exist_ok=True)

    f = ROOT.TFile(args.file)
    events = f.EvTree

    ntotal = events.GetEntries()
    nend = min(ntotal, args.num_events) if args.num_events > 0 else ntotal
    nbeg = args.start_event

    # CerB3 is problematic
    channels = [
            'CerA0', 'CerA1', 'CerA2', 'CerA3',
            'CerB0', 'CerB1', 'CerB2', # 'CerB3',
            'CerC0', 'CerC1', 'CerC2', 'CerC3',
            'CerD0', 'CerD1', 'CerD2', 'CerD3',
            ]

    data = []
    for iev in np.arange(nbeg, nend):
        if iev % 100000 == 0:
            print('{:d}/{:d}'.format(iev, nend - nbeg))
        events.GetEntry(iev)
        datum = [max_peak_height(getattr(events, chan)) for chan in channels]
        nch = np.sum(np.array(datum) > args.signal_cut)
        sigsum = np.sum(datum)
        if nch > 0:
            datum += [iev, nch, sigsum]
            data.append(datum)
    df = pd.DataFrame(data=data, columns=channels + ['event', 'nch', 'sig_sum'])
    df.to_csv('test.csv')
    print(df)

