import ROOT
import os
import json
import numpy as np
import pandas as pd
import argparse
from matplotlib import pyplot as plt
from ROOT import gROOT, gSystem, gInterpreter, gStyle, gPad
from plot_utils import prepare_canvas, prepare_figure, my_style, get_hist_contents, nthreads

ROOT.ROOT.EnableImplicitMT(nthreads)

# make sure the relative path for root loading is correct
owd = os.getcwd()
script_dir = os.path.dirname(os.path.realpath(__file__))
os.chdir(os.path.join(script_dir, '..'))
print('working directory is ' + os.getcwd())

# batch mode
gROOT.SetBatch(True)
# rootlogon macro
gROOT.Macro('rootlogon.C')

# the functions that process data are defined in scripts/root/dataframe_analysis.C
func_code = open(os.path.join(script_dir, 'root', 'dataframe_analysis.C')).read()
gInterpreter.Declare(func_code)

# global style
my_style.cd()

# argument parser
parser = argparse.ArgumentParser()
parser.add_argument('json_command', help='json file of plot commands')
parser.add_argument('-r', dest='run', type=int, help='run number', required=True)
parser.add_argument('-s', dest='split', default=0, type=int, help='split file number')
parser.add_argument('-o', '--plot-dir', dest='out_dir', default='./plots', help='output directory')
parser.add_argument('--layout', dest='layout', default=os.path.join(script_dir, '..', 'database/channels_layout.json'),
                    help='json data for channels layout')
args = parser.parse_args()

# recover paths
for attr in ['json_command', 'out_dir', 'layout']:
    setattr(args, attr, os.path.join(owd, getattr(args, attr)))

os.makedirs(args.out_dir, exist_ok=True)

# read json files
with open(args.json_command) as f:
    plot_data = json.load(f)

with open(args.layout, 'r') as f:
    channels_layout = json.load(f)


# get histogram from dataframe
def get_hist(root_df, fcode, bins, xmin, xmax, filcode):
    df_temp = root_df.Define('my_temp', fcode)
    if filcode:
        df_temp = df_temp.Filter(filcode)
    return df_temp.Histo1D(ROOT.RDF.TH1DModel('my_temp', '', bins, xmin, xmax), 'my_temp')


# plot with matplotlib
def mat_plots(axs, path, func_code, hist_range, filter_code='', refs={}, fontsize=18, normalized=False,
              yscale='linear', plot_kw={}, **kwargs):
    print(os.path.join(owd, path.format(run=args.run, split=args.split)))
    df = ROOT.RDataFrame('EvTree', os.path.join(owd, path.format(run=args.run, split=args.split)))
    bprops = dict(boxstyle='round', facecolor='wheat', alpha=0.1)

    for ch, ax in axs.items():
        # df = df.Define('{}_mpeak'.format(ch), 'get_maximum_peak({}.peaks)'.format(ch))
        kw = refs.get(ch, {})
        fcode = func_code.format(ch=ch, **kw).replace('[', '').replace(']', '')
        filcode = filter_code.format(ch=ch, **kw).replace('[', '').replace(']', '')
        x, y = get_hist_contents( get_hist(df, fcode, *hist_range, filcode) )
        wts = y/np.sum(y) if normalized else y
        ax.set_yscale(yscale)
        ax.hist(x, bins=np.linspace(*hist_range[1:], hist_range[0] + 1), weights=wts, **plot_kw)
        ax.text(0.95, 0.95, ch, transform=ax.transAxes, fontsize=fontsize, bbox=bprops,
                verticalalignment='top', horizontalalignment='right')


# generate plots
for group, data in plot_data.items():
    # get default layout if no channels provided
    channels = data.get('channels', channels_layout.get(group, {}))
    refs = data.get('refs', {})
    # path
    if isinstance(refs, str):
        cmd_dir = os.path.dirname(os.path.realpath(args.json_command))
        with open(os.path.join(cmd_dir, refs.format(run=args.run, split=args.split)), 'r') as f:
            refs = json.load(f)

    for kwargs in data['plots']:
        fig = plt.figure(figsize=(16, 9), dpi=160)
        axs = prepare_figure(fig, channels, **kwargs)
        for sub_kw in kwargs['subs']:
            skw = kwargs.copy()
            skw.update(sub_kw)
            pkw = {k: v for k, v in sub_kw.items() if k not in ['path', 'func_code', 'filter_code']}
            mat_plots(axs, refs=refs, plot_kw=pkw, **skw)
        fig.savefig(os.path.join(args.out_dir, kwargs['save']))

