import argparse
import os
import json
import ROOT
import random
import numpy as np
import pandas as pd
from ROOT import gROOT, gInterpreter
from plot_utils import prepare_canvas, my_style, nthreads, get_hist_contents2

ROOT.ROOT.EnableImplicitMT(nthreads)

if __name__ == "__main__":
    # batch mode
    gROOT.SetBatch(True)

    # make sure the relative path for root loading is correct
    owd = os.getcwd()
    script_dir = os.path.dirname(os.path.realpath(__file__))
    os.chdir(os.path.join(script_dir, '..'))
    print('working directory is ' + os.getcwd())

    # signal layouts
    layout_json = os.path.join(script_dir, '..', 'database', 'channels_layout.json')
    with open(layout_json, 'r') as f:
        channels_layout = json.load(f)
    signal_types = {
        'LAPPD': [
            ('Pixels', 'lappd_all', [1, 2, 3, 9, 10, 11, 19, 20, 21], 'Peak Height (mV)', (160, 0, 15), 81.92)
        ],
        'MAPMT': [
            ('PMTs', 'mapmt_sums_all', np.arange(2, 10), 'Est. Number of Photo-electrons', (160, 0, 60), 80),
            ('Quads', 'mapmt_quads_all', np.arange(3, 20), 'Est. Number of Photo-electrons', (160, 0, 60), 130),
            # ('Quads', 'mapmt_quads_all', [1, 2, 3, 9, 10, 11, 19, 20, 21], 'Peak Height (mV)', (160, 0, 120), 81.92),
        ]
    }

    # argument parser
    parser = argparse.ArgumentParser('draw events')
    parser.add_argument('stype', help='signal type, supporting: {}'.format(list(signal_types)))
    parser.add_argument('root_file', help='a root file of waveform data')
    parser.add_argument('-o', '--plot-dir', dest='out_dir', default='./plots', help='output directory')
    parser.add_argument('--csv2d', dest='csv2d', default='', help='path to save 2d histogram data')
    args = parser.parse_args()

    if args.stype not in signal_types:
        print('signal type must be in {}'.format(list(signal_types)))
        exit(-1)

    # recover paths
    for attr in ['root_file', 'out_dir']:
        setattr(args, attr, os.path.join(owd, getattr(args, attr)))

    os.makedirs(args.out_dir, exist_ok=True)

    # root macros
    gROOT.Macro('rootlogon.C')
    # declare root graph code
    func_code = open(os.path.join(script_dir, 'root', 'dataframe_analysis.C')).read()
    gInterpreter.Declare(func_code)

    # style
    my_style.cd()

    # root dataframe
    rdf = ROOT.RDataFrame('EvTree', args.root_file)
    df = pd.DataFrame()
    for name, signal, nhits, x_label, x_hrange, spe in signal_types[args.stype]:
        nh_label = 'N_{}'.format(name)
        sname = '{} {}'.format(args.stype.replace('_', ' '), name)
        sigch = '{}_height'.format(signal)
        rdf = rdf.Define(sigch, '{}.height * {}'.format(signal, 1./spe))

        c = ROOT.TCanvas('signal_sum', '', 1920, 1080)
        title = '{} Signal Sums'.format(sname)
        pads = prepare_canvas(c, [[nh] for nh in nhits], title, x_label, 'Counts')
        hists = []
        for i, nh in enumerate(nhits):
            pads[nh].cd()
            hist = rdf.Filter('{}.pos == {}'.format(signal, nh))\
                      .Histo1D(ROOT.RDF.TH1DModel('hsum{}'.format(nh), '{} = {}'.format(nh_label, nh), *x_hrange),
                               sigch)
            hist.SetLineColor(1)
            hist.SetFillColorAlpha(i + 1, 0.5)
            hist.Draw('hist & same')
            hists.append(hist)
        c.SaveAs(os.path.join(args.out_dir, '{}_sum_nhits.png'.format(sname.replace(' ', '_').lower())))

        # 2d hist
        min_hit = min(nhits)
        c = ROOT.TCanvas('signal_dist', '', 1440, 1080)
        pads = prepare_canvas(c, [[0, 1]], '', x_label, nh_label)
        minh, maxh = 0.5, max(nhits) + 0.5
        filter_code = ['true', '{}.pos >= {}'.format(signal, min_hit)]
        for i, fc in enumerate(filter_code):
            pads[i].cd()
            pads[i].SetRightMargin(0.13)
            pads[i].SetBottomMargin(0.08)
            hist = rdf.Filter(fc)\
                      .Histo2D(ROOT.RDF.TH2DModel('hsum2d', '', *x_hrange, int(maxh - minh), minh, maxh),
                               sigch, '{}.pos'.format(signal))
            hist.DrawClone('colz')
        c.SaveAs(os.path.join(args.out_dir, '{}_sum_2d.png'.format(sname.replace(' ', '_').lower())))
        # hist data
        x, y, counts = get_hist_contents2(hist)
        tmp = pd.DataFrame(columns=['npe', 'nch', 'counts'], data=np.vstack([x.flatten(), y.flatten(), counts.flatten()]).T)
        tmp.loc[:, 'type'] = name.lower()
        df = pd.concat([df, tmp])
        # without any cut
        hist2 = rdf.Histo2D(ROOT.RDF.TH2DModel('hsum2d2', '', *x_hrange, int(maxh - minh), minh, maxh), sigch, '{}.pos'.format(signal))
        x, y, counts = get_hist_contents2(hist2)
        tmp = pd.DataFrame(columns=['npe', 'nch', 'counts'], data=np.vstack([x.flatten(), y.flatten(), counts.flatten()]).T)
        tmp.loc[:, 'type'] = name.lower() + '_nocut'
        df = pd.concat([df, tmp])


        c = ROOT.TCanvas('total_signal', '', 1920, 1080)
        title = '{} Signal Sum ({} >= {})'.format(sname, nh_label, min_hit)
        pads = prepare_canvas(c, [[0]], title, x_label, 'Normalized Counts')
        pads[0].cd()
        pads[0].SetLogy()
        hist = rdf.Filter('{}.pos >= {}'.format(signal, min_hit))\
                  .Histo1D(ROOT.RDF.TH1DModel('hsum', '', *x_hrange), sigch)
        hist.SetLineColor(1)
        hist.SetFillColorAlpha(38, 0.5)
        chist = hist.DrawNormalized('hist & same')
        leg = ROOT.TLegend(0.5, 0.8, 0.9, 0.9)
        leg.SetHeader('Total Entries {}'.format(rdf.Count().GetValue()), 'C')
        leg.Draw()
        c.SaveAs(os.path.join(args.out_dir, '{}_sum.png'.format(sname.replace(' ', '_').lower())))
    # save data
    if args.csv2d:
        df.to_csv(args.csv2d, index=False)
