// a script for analyze the data with root dataframe

#include "Fadc250Data.h"


using namespace fdec;
using namespace ROOT::VecOps;

// for reducing the data size
Fadc250Data get_peaks(const Fadc250Data &data)
{
    Fadc250Data res;
    res.peaks = data.peaks;
    return res;
}

std::vector<double> unpack_time(const std::vector<Peak> &peaks, bool cut_overflow = true, double ref = 0., double scale = 1.)
{
    std::vector<double> res;
    for (auto &p : peaks) { if (!p.overflow || !cut_overflow) res.push_back(scale*(p.time - ref)); }
    return res;
}

std::vector<double> unpack_time(const RVec<Peak> &peaks, bool cut_overflow = true, double ref = 0., double scale = 1.)
{
    std::vector<double> res;
    for (auto &p : peaks) { if (!p.overflow || !cut_overflow) res.push_back(scale*(p.time - ref)); }
    return res;
}

std::vector<double> unpack_height(const std::vector<Peak> &peaks, bool cut_overflow = true, double ref = 0., double scale = 1.)
{
    std::vector<double> res;
    for (auto &p : peaks) { if (!p.overflow || !cut_overflow) res.push_back(scale*(p.height - ref)); }
    return res;
}

std::vector<double> unpack_height(const RVec<Peak> &peaks, bool cut_overflow = true, double ref = 0., double scale = 1.)
{
    std::vector<double> res;
    for (auto &p : peaks) { if (!p.overflow || !cut_overflow) res.push_back(scale*(p.height - ref)); }
    return res;
}

std::vector<double> unpack_integral(const std::vector<Peak> &peaks, bool cut_overflow = true, double ref = 0., double scale = 1.)
{
    std::vector<double> res;
    for (auto &p : peaks) { if (!p.overflow || !cut_overflow) res.push_back(scale*(p.integral - ref)); }
    return res;
}

std::vector<double> unpack_integral(const RVec<Peak> &peaks, bool cut_overflow = true, double ref = 0., double scale = 1.)
{
    std::vector<double> res;
    for (auto &p : peaks) { if (!p.overflow || !cut_overflow) res.push_back(scale*(p.integral - ref)); }
    return res;
}

// get peak in time window
bool time_cut(const Peak &p, double tmin, double tmax)
{
    return (p.time < tmax) && (p.time >= tmin);
}

bool time_cut_rel(const Peak &p, double tmin, double tmax, double ref)
{
    return ((p.time - ref) < tmax) && ((p.time - ref) >= tmin);
}

Peak get_maximum_peak(const std::vector<Peak> &peaks, bool cut_overflow = true)
{
    Peak res(0., 0., 0.);

    for (size_t i = 0; i < peaks.size(); ++i) {
        auto &p = peaks[i];
        if ((!p.overflow || !cut_overflow) && p.height > res.height) {
            res = p;
            res.pos = i;
        }
    }
    return res;
}

// get the maximum peak with a filter
template<class Filter, typename... Args>
Peak get_maximum_peak(const std::vector<Peak> &peaks, Filter &&f, Args&&... args)
{
    Peak res(0., 0., 0.);

    for (size_t i = 0; i < peaks.size(); ++i) {
        auto &p = peaks[i];
        if (!p.overflow && f(p, args...) && p.height > res.height) {
            res = p;
            res.pos = i;
        }
    }

    return res;
}

// get the maximum peak with a filter
template<class Filter, typename... Args>
Peak get_maximum_peak(const RVec<Peak> &peaks, Filter &&f, Args&&... args)
{
    Peak res(0., 0., 0.);

    for (size_t i = 0; i < peaks.size(); ++i) {
        auto &p = peaks[i];
        if (!p.overflow && f(p, args...) && p.height > res.height) {
            res = p;
            res.pos = i;
        }
    }

    return res;
}

Peak peak_sum(const std::vector<Peak> &peaks, double thres = 0.)
{
    Peak res(0., 0., 0., 0);

    for (auto &peak : peaks) {
        if (peak.height > thres) {
            res.time += peak.time;
            res.integral += peak.integral;
            res.height += peak.height;
            res.pos ++;
        }
    }
    res.time /= (double)res.pos;
    return res;
}

// get the number of peaks following the first one
int count_following_peaks(const std::vector<Peak> &peaks, double time_window = 30)
{
    if (peaks.empty()) return 0;

    int count = 0;
    for (size_t i = 1; i < peaks.size(); ++i) {
        if (peaks[i].time - peaks[0].time < time_window) count ++;
    }
    return count;
}

double integrate_raw(const Fadc250Data &data, int start = 10, int end = 21, double scale = 1.)
{
    double res = 0.;
    if (data.raw.size() == 0) return res;
    for (int i = start; i < data.raw.size() && i < end; ++i) {
        res += data.raw[i] - data.ped.mean;
    }
    return scale*res;
}

int count_peaks(const std::vector<Peak> &peaks, int start = -1, int end = -1, double thres = 100.)
{
    int count = 0;
    for (auto &p : peaks) {
        if ((start < 0 || p.pos >= start) && (end < 0 || p.pos <= end)) {
            count += int(p.height > thres);
        }
    }
    return count;
}

bool cher_cut_rough(const std::vector<Fadc250Data> &chans, double thres = 50.,
                    int cmin = 7, int cmax = 11, double smin = 600, double smax = 1000)
{
    int count = 0;
    double sum = 0.;
    for (auto &ch : chans) {
        if (ch.peaks.empty()) continue;
        auto peak = get_maximum_peak(ch.peaks, time_cut, 20, 50);
        if (peak.height < thres) continue;

        count ++;
        sum += peak.height;
    }
    return count > 4;
    // return count <= cmax && count >= cmin && sum <= smax && sum >= smin;
}

